# https://github.com/JulienPalard/Pipe/blob/main/pipe.py
import functools


class PipeFirst:
    """
    Takes the input into `|` and passes to the first argument of the function

    Described as:
    >>> first = Pipe(lambda iterable: next(iter(iterable)))
    and used as:
    >>> print([1, 2, 3] | first)
    1

    Or represent a Pipeable Function :
    It's a function returning a Pipe
    Described as :
    >>> select = Pipe(lambda iterable, pred: (pred(x) for x in iterable))
    and used as:
    >>> print([1, 2, 3] | select(lambda x: x * 2))
    2, 4, 6
    """

    def __init__(self, function):
        self.function = function
        functools.update_wrapper(self, function)

    def __ror__(self, other):
        return self.function(other)

    def __call__(self, *args, **kwargs):
        return PipeFirst(
            lambda iterable, *args2, **kwargs2: self.function(
                iterable, *args, *args2, **kwargs, **kwargs2
            )
        )


class PipeSecond:
    """
    Represent a Pipeable Element :
    Described as :
    first = Pipe(lambda iterable: next(iter(iterable)))
    and used as :
    print [1, 2, 3] | first
    printing 1

    Or represent a Pipeable Function :
    It's a function returning a Pipe
    Described as :
    select = Pipe(lambda iterable, pred: (pred(x) for x in iterable))
    and used as :
    print [1, 2, 3] | select(lambda x: x * 2)
    # 2, 4, 6
    """

    def __init__(self, function):
        self.function = function
        functools.update_wrapper(self, function)

    def __ror__(self, other):
        return self.function(other)

    def __call__(self, *args, **kwargs):
        return PipeSecond(
            lambda iterable, *args2, **kwargs2: self.function(
                *args, *args2, iterable, **kwargs, **kwargs2
            )
        )



from operator import itemgetter
import random

import funcy as f
import funcy_pipe as fp
from funcy import none
from funcy_pipe.pipe import PipeFirst, PipeSecond


# https://github.com/Suor/funcy/pull/143
def where_not(mappings, **cond):
    """Iterates over mappings containing all pairs not in cond."""
    items = cond.items()
    match = lambda m: none(k in m and m[k] == v for k, v in items)
    return filter(match, mappings)


fp.where_not = PipeFirst(where_not)


def where_not_attr(objects, **cond):
    items = cond.items()
    match = lambda obj: none(hasattr(obj, k) and getattr(obj, k) == v for k, v in items)
    return filter(match, objects)


fp.where_not_attr = PipeFirst(where_not_attr)


# https://github.com/Suor/funcy/pull/140
def where_attr(objects, **cond):
    items = cond.items()
    match = lambda obj: all(hasattr(obj, k) and getattr(obj, k) == v for k, v in items)
    return filter(match, objects)


fp.where_attr = PipeFirst(where_attr)


# NOTE this replaces an existing method
# https://github.com/Suor/funcy/pull/142
def pluck(key, mappings):
    """Iterates over values for key, or multiple keys, in mappings."""
    if isinstance(key, (list, tuple)):
        return map(itemgetter(*key), mappings)
    else:
        return map(itemgetter(key), mappings)


fp.pluck = PipeSecond(pluck)


# https://github.com/Suor/funcy/pull/147
def shuffled(seq):
    new_seq = list(seq)
    random.shuffle(new_seq)
    return new_seq


fp.shuffled = PipeFirst(shuffled)


# TODO propose as funcy addition
def join_str(sep, seq):
    return sep.join(seq)


fp.join_str = PipeSecond(join_str)


# TODO propose as funcy addition
def sort(iterable, key=None, reverse=False):
    # TODO what about objects?
    # if key is str, assume it is a dict key
    if key is not None and type(key) is str:
        key = f.rpartial(f.get_in, [key])

    return sorted(iterable, key=key, reverse=reverse)


fp.sort = PipeFirst(sort)


def patch():
    f.where_attr = where_attr
    f.where_not = where_not
    f.where_not_attr = where_not_attr
    f.shuffled = shuffled
    f.pluck = pluck
    f.join_str = join_str
    f.sort = sort



import functools
import funcy

from .funcy_extensions import patch
from .pipe import PipeFirst, PipeSecond

__all__ = []

# where the first param is the iterable
PIPE_FIRST_EXCEPTIONS = ["omit", "iteritems", "itervalues", "empty"]
PIPE_FIRST_OMISSIONS = ["partial", "rpartial", "curry", "rcurry"]


def export(func):
    globals()["__all__"].append(func.__name__)
    return func


export(patch)


def apply_decorator_and_export(module, decorator):
    decorated_functions = {}

    for name in dir(module):
        if name.startswith("__"):
            continue

        obj = getattr(module, name)

        if not callable(obj):
            continue

        if name in PIPE_FIRST_OMISSIONS:
            globals()[name] = obj
            __all__.append(name)
            continue

        decorated_functions[name] = (
            # TODO lol why pass in decorator if this logic exists?
            PipeFirst(obj)
            if name in PIPE_FIRST_EXCEPTIONS
            else decorator(obj)
        )

        # add function to the module
        globals()[name] = decorated_functions[name]
        __all__.append(name)

    return decorated_functions


fp = apply_decorator_and_export(funcy, PipeSecond)


@PipeFirst
@export
def to_list(iterable):
    return list(iterable)


@PipeFirst
@export
def log(iterable):
    # TODO maybe use pretty print
    print(iterable)
    return iterable


# breakpoint, avoid collision with built-in breakpoint
@PipeFirst
@export
def bp(iterable):
    breakpoint()
    return iterable


# TODO there's `first` but it doesn't throw an exception if there's more than one
@export
def exactly_one(comprehension):
    # TODO detect if we are working with a generator/iterator and maybe use first?

    if len(comprehension) != 1:
        raise Exception("Expected to find exactly one matching item")

    return comprehension[0]


# TODO should document in funcy and remove this
# https://github.com/Suor/funcy/commit/bbc249672df3839fc0e3f3fc9fbb4483978886b5
@PipeFirst
@export
def reduce(iterable, func, initial):
    return functools.reduce(func, iterable, initial)


@PipeFirst
@export
def pmap(iterable, func):
    """
    >>> [1, 2, 3] | fp.pmap(fp.omit("key"))
    """
    return iterable | fp["map"](lambda x: x | func)
